from typing import Optional, Tuple, Any
import platform
import struct
from functools import lru_cache
import pathlib
import os

LEFT_ARROW                             = " \u2190 "
RIGHT_ARROW                            = " \u2192 "
DOWN_ARROW                             = "\u21b3"
HORIZONTAL_LINE                        = "\u2500"
VERTICAL_LINE                          = "\u2502"
CROSS                                  = "\u2718 "
TICK                                   = "\u2713 "
ANSI_SPLIT_RE                          = r"(\033\[[\d;]*m)"


class Color:
    """Used to colorify terminal output."""
    colors = {
        "normal": "\033[0m",
        "gray": "\033[1;38;5;240m",
        "light_gray": "\033[0;37m",
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "pink": "\033[35m",
        "cyan": "\033[36m",
        "bold": "\033[1m",
        "underline": "\033[4m",
        "underline_off": "\033[24m",
        "highlight": "\033[3m",
        "highlight_off": "\033[23m",
        "blink": "\033[5m",
        "blink_off": "\033[25m",
    }

    @staticmethod
    def redify(msg: str) -> str:        return Color.colorify(msg, "red")
    @staticmethod
    def greenify(msg: str) -> str:      return Color.colorify(msg, "green")
    @staticmethod
    def blueify(msg: str) -> str:       return Color.colorify(msg, "blue")
    @staticmethod
    def yellowify(msg: str) -> str:     return Color.colorify(msg, "yellow")
    @staticmethod
    def grayify(msg: str) -> str:       return Color.colorify(msg, "gray")
    @staticmethod
    def light_grayify(msg: str) -> str: return Color.colorify(msg, "light_gray")
    @staticmethod
    def pinkify(msg: str) -> str:       return Color.colorify(msg, "pink")
    @staticmethod
    def cyanify(msg: str) -> str:       return Color.colorify(msg, "cyan")
    @staticmethod
    def boldify(msg: str) -> str:       return Color.colorify(msg, "bold")
    @staticmethod
    def underlinify(msg: str) -> str:   return Color.colorify(msg, "underline")
    @staticmethod
    def highlightify(msg: str) -> str:  return Color.colorify(msg, "highlight")
    @staticmethod
    def blinkify(msg: str) -> str:      return Color.colorify(msg, "blink")

    @staticmethod
    def colorify(text: str, attrs: str) -> str:
        """Color text according to the given attributes."""

        colors = Color.colors
        msg = [colors[attr] for attr in attrs.split() if attr in colors]
        msg.append(str(text))
        if colors["highlight"] in msg:   msg.append(colors["highlight_off"])
        if colors["underline"] in msg:   msg.append(colors["underline_off"])
        if colors["blink"] in msg:       msg.append(colors["blink_off"])
        msg.append(colors["normal"])
        return "".join(msg)


def get_terminal_size() -> Tuple[int, int]:
    """Return the current terminal size."""
    if platform.system() == "Windows":
        from ctypes import windll, create_string_buffer
        hStdErr = -12
        herr = windll.kernel32.GetStdHandle(hStdErr)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(herr, csbi)
        if res:
            _, _, _, _, _, left, top, right, bottom, _, _ = struct.unpack("hhhhHhhhhhh", csbi.raw)
            tty_columns = right - left + 1
            tty_rows = bottom - top + 1
            return tty_rows, tty_columns
        else:
            return 600, 100
    else:
        import fcntl
        import termios
        try:
            tty_rows, tty_columns = struct.unpack("hh", fcntl.ioctl(1, termios.TIOCGWINSZ, "1234"))
            return tty_rows, tty_columns
        except OSError:
            return 600, 100


def pprint(*args: str, end="\n", sep=" ", **kwargs: Any) -> None:
    """Wrapper around print(), using string buffering feature."""
    parts = args
    print(*parts, sep=sep, end=end, **kwargs)
    return


def titlify(text: str, color: Optional[str] = None, msg_color: Optional[str] = None) -> str:
    """Print a centered title."""
    cols = get_terminal_size()[1]
    nb = (cols - len(text) - 2) // 2
    if color is None:
        color = "gray"
    if msg_color is None:
        msg_color = "cyan"

    msg = [Color.colorify(f"{HORIZONTAL_LINE * nb} ", color),
           Color.colorify(text, msg_color),
           Color.colorify(f" {HORIZONTAL_LINE * nb}", color)]
    return "".join(msg)


def err(msg: str) -> None:
    pprint(f"{Color.colorify('[!]', 'bold red')} {msg}")


def warn(msg: str) -> None:
    pprint(f"{Color.colorify('[*]', 'bold yellow')} {msg}")


def ok(msg: str) -> None:
    pprint(f"{Color.colorify('[+]', 'bold green')} {msg}")


def info(msg: str) -> None:
    pprint(f"{Color.colorify('[+]', 'bold blue')} {msg}")


def gef_pystring(x: bytes) -> str:
    """Returns a sanitized version as string of the bytes list given in input."""
    res = str(x, encoding="utf-8")
    substs = [("\n", "\\n"), ("\r", "\\r"), ("\t", "\\t"), ("\v", "\\v"), ("\b", "\\b"), ]
    for x, y in substs: res = res.replace(x, y)
    return res


#
# ██████╗ ███████╗ ██████╗ ██████╗ ███╗   ███╗██████╗ ██████╗ ██████╗ ██████╗  ██████╗
# ██╔══██╗██╔════╝██╔════╝██╔═══██╗████╗ ████║██╔══██╗╚════██╗██╔══██╗██╔══██╗██╔════╝
# ██║  ██║█████╗  ██║     ██║   ██║██╔████╔██║██████╔╝ █████╔╝██║  ██║██████╔╝██║  ███╗
# ██║  ██║██╔══╝  ██║     ██║   ██║██║╚██╔╝██║██╔═══╝ ██╔═══╝ ██║  ██║██╔══██╗██║   ██║
# ██████╔╝███████╗╚██████╗╚██████╔╝██║ ╚═╝ ██║██║     ███████╗██████╔╝██████╔╝╚██████╔╝
# ╚═════╝ ╚══════╝ ╚═════╝ ╚═════╝ ╚═╝     ╚═╝╚═╝     ╚══════╝╚═════╝ ╚═════╝  ╚═════╝
#                            angr-management server
#
# by clasm, 2021.
#

import threading
import traceback

from PySide2.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QLineEdit, QMessageBox, QGridLayout
from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.workspace import Workspace

from .server import AngrDecompilerServer

#
# UI
#


class ConfigDialog(QDialog):
    def __init__(self, instance, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Configure Decomp2GEF")
        self._main_layout = QVBoxLayout()
        self._instance = instance
        self._host_edit = None  # type:QLineEdit
        self._port_edit = None  # type:QLineEdit

        self._init_widgets()
        self.setLayout(self._main_layout)
        self.show()

    def _init_widgets(self):
        upper_layout = QGridLayout()

        host_label = QLabel(self)
        host_label.setText("Host")
        self._host_edit = QLineEdit(self)
        self._host_edit.setText("localhost")
        row = 0
        upper_layout.addWidget(host_label, row, 0)
        upper_layout.addWidget(self._host_edit, row, 1)
        row += 1

        port_label = QLabel(self)
        port_label.setText("Port")
        self._port_edit = QLineEdit(self)
        self._port_edit.setText("3662")
        upper_layout.addWidget(port_label, row, 0)
        upper_layout.addWidget(self._port_edit, row, 1)
        row += 1

        # buttons
        self._ok_button = QPushButton(self)
        self._ok_button.setText("OK")
        self._ok_button.setDefault(True)
        self._ok_button.clicked.connect(self._on_ok_clicked)
        cancel_button = QPushButton(self)
        cancel_button.setText("Cancel")
        cancel_button.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(cancel_button)

        # main layout
        self._main_layout.addLayout(upper_layout)
        self._main_layout.addLayout(buttons_layout)

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        host = self._host_edit.text()
        port = self._port_edit.text()

        if not host:
            QMessageBox(self).critical(None, "Invalid host",
                                       "Host cannot be empty."
                                       )
            return

        if not port:
            QMessageBox(self).critical(None, "Invalid port",
                                       "Port cannot be empty"
                                       )
            return

        decomp_server = AngrDecompilerServer(self._instance)
        t = threading.Thread(target=decomp_server.start_xmlrpc_server, kwargs={"host": host, "port": int(port, 0)})
        t.daemon = True
        try:
            t.start()
        except Exception as e:
            QMessageBox(self).critical(None, "Error starting Decomp2GEF Server", str(e))
            traceback.print_exc()
            return

        self.close()

    def _on_cancel_clicked(self):
        self.close()


class Decomp2DbgPlugin(BasePlugin):
    def __init__(self, workspace: Workspace):
        """
        The entry point for the Decomp2Gef plugin.
        @param workspace:   an AM _workspace (usually found in _instance)
        """
        super().__init__(workspace)
        self._workspace = workspace
        self._instance = workspace.instance

    MENU_BUTTONS = ('Configure Decomp2GEF...', )
    MENU_CONFIG_ID = 0

    def handle_click_menu(self, idx):
        # sanity check on menu selection
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.instance.project.am_none:
            return

        mapping = {
            self.MENU_CONFIG_ID: self.open_config_dialog,
        }

        # call option mapped to each menu pos
        mapping.get(idx)()

    def open_config_dialog(self):
        if self.workspace.instance.project.am_none:
            # project does not exist yet
            return

        config = ConfigDialog(self._instance)
        config.exec_()



import pathlib
from typing import Dict, Optional, Sequence, Union, Any, List, Callable
import subprocess
import os
from functools import lru_cache
import functools
import collections
import re
import tempfile
import hashlib

from elftools.elf.elffile import ELFFile

from ...utils import gef_pystring, warn, err

import gdb

GLOBAL_TMP_DIR = os.path.join(tempfile.gettempdir(), "d2d")
ARCH = None


def identify_arch():
    global ARCH
    if ARCH:
        return ARCH

    with open(get_filepath(), "rb") as fp:
        elf = ELFFile(fp)

    ARCH = elf.get_machine_arch()
    return ARCH


def get_arch_func_args():
    # args taken from GEF
    arch_args = {
        "x64": ["$rdi", "$rsi", "$rdx", "$rcx", "$r8", "$r9"],
        "x86": [f'$esp+{x}' for x in range(0, 28, 4)],
        "ARM": ["$r0", "$r1", "$r2", "$r3"],
        "SPARC": ["$o0 ", "$o1 ", "$o2 ", "$o3 ", "$o4 ", "$o5 ", "$o7 "],
        "MIPS": ["$a0", "$a1", "$a2", "$a3"],
        "RISC-V": ["$a0", "$a1", "$a2", "$a3", "$a4", "$a5", "$a6", "$a7"]
    }

    arch = identify_arch()
    try:
        args = arch_args[arch]
    except KeyError:
        args = []

    return args


def vmmap_base_addrs():
    addr_maps = {}
    mappings = gdb.execute("info proc mappings", to_string=True).split("\n")
    for mapping in mappings:
        try:
            addr = int(re.findall(r"0x[0-9a-fA-F]+", mapping)[0], 16)
            path = mapping.split(" ")[-1]
        except IndexError:
            continue

        # always use the lowest addr
        if path in addr_maps or path.startswith("["):
            continue

        if addr and path:
            addr_maps[path] = addr

    return addr_maps


def find_text_segment_base_addr(is_remote=False):
    with open(get_filepath(), 'rb') as fp:
        binary_hash = hashlib.md5(fp.read()).hexdigest()

    if is_remote:
        def _should_hash_cmp(path_name):
            return True
    else:
        def _should_hash_cmp(path_name):
            return path_name == get_filepath()

    for path, addr in vmmap_base_addrs().items():
        if _should_hash_cmp(path):
            file = download_file(path) if is_remote else path
            with open(file, 'rb') as fp:
                other_file_hash = hashlib.md5(fp.read()).hexdigest()

            if other_file_hash == binary_hash:
                return addr
    else:
        raise Exception("Unable to find the text segment base addr, please report this!")


@lru_cache()
def is_32bit():
    ptr_size = int(gdb.execute("p sizeof(long long)", to_string=True).split("= ")[1].strip(), 0)
    if ptr_size == 4:
        return True

    return False


def pc():
    try:
        pc_ = int(gdb.execute("print/x $pc",to_string=True).split(" ")[-1], 16)
    except Exception:
        pc_ = None

    return pc_


#
# GEF Clone
#

@lru_cache()
def which(program: str) -> Optional[pathlib.Path]:
    """Locate a command on the filesystem."""
    for path in os.environ["PATH"].split(os.pathsep):
        dirname = pathlib.Path(path)
        fpath = dirname / program
        if os.access(fpath, os.X_OK):
            return fpath

    raise FileNotFoundError(f"Missing file `{program}`")

def exec_external(command: Sequence[str], as_list: bool = False, **kwargs: Any) -> Union[str, List[str]]:
    """Execute an external command and return the result."""
    res = subprocess.check_output(command, stderr=subprocess.STDOUT, shell=kwargs.get("shell", False))
    return [gef_pystring(_) for _ in res.splitlines()] if as_list else gef_pystring(res)

@lru_cache(32)
def checksec(filename: str) -> Dict[str, bool]:
    """Check the security property of the ELF binary. The following properties are:
    - Canary
    - NX
    - PIE
    - Fortify
    - Partial/Full RelRO.
    Return a dict() with the different keys mentioned above, and the boolean
    associated whether the protection was found."""
    readelf = which("readelf")

    def __check_security_property(opt: str, filename: str, pattern: str) -> bool:
        cmd   = [readelf,]
        cmd  += opt.split()
        cmd  += [filename,]
        lines = exec_external(cmd, as_list=True)
        for line in lines:
            if re.search(pattern, line):
                return True
        return False

    results = collections.OrderedDict()
    results["Canary"] = __check_security_property("-rs", filename, r"__stack_chk_fail") is True
    has_gnu_stack = __check_security_property("-W -l", filename, r"GNU_STACK") is True
    if has_gnu_stack:
        results["NX"] = __check_security_property("-W -l", filename, r"GNU_STACK.*RWE") is False
    else:
        results["NX"] = False
    results["PIE"] = __check_security_property("-h", filename, r":.*EXEC") is False
    results["Fortify"] = __check_security_property("-s", filename, r"_chk@GLIBC") is True
    results["Partial RelRO"] = __check_security_property("-l", filename, r"GNU_RELRO") is True
    results["Full RelRO"] = results["Partial RelRO"] and __check_security_property("-d", filename, r"BIND_NOW") is True
    return results



@lru_cache()
def is_remote_debug() -> bool:
    """"Return True is the current debugging session is running through GDB remote session."""
    return "remote" in gdb.execute("maintenance print target-stack", to_string=True)


def pid() -> int:
    """Return the PID of the target process."""
    pid_ = gdb.selected_inferior().pid
    if not pid_:
        pid_ = gdb.selected_thread().ptid[1]
        if not pid_:
            raise RuntimeError("cannot retrieve PID for target process")

    return pid_



@lru_cache()
def get_filepath() -> Optional[str]:
    """Return the local absolute path of the file currently debugged."""
    filename = gdb.current_progspace().filename

    if is_remote_debug():
        # if no filename specified, try downloading target from /proc
        if filename is None:
            pid_ = pid()
            if pid_ > 0:
                return download_file(f"/proc/{pid_:d}/exe", use_cache=True)
            return None

        # if target is remote file, download
        elif filename.startswith("target:"):
            fname = filename[len("target:") :]
            return download_file(fname, use_cache=True, local_name=fname)

        elif filename.startswith(".gnu_debugdata for target:"):
            fname = filename[len(".gnu_debugdata for target:") :]
            return download_file(fname, use_cache=True, local_name=fname)

        return filename
    else:
        if filename is not None:
            return filename
        # inferior probably did not have name, extract cmdline from info proc
        return get_path_from_info_proc()


def get_path_from_info_proc() -> Optional[str]:
    for x in gdb.execute("info proc", to_string=True).splitlines():
        if x.startswith("exe = "):
            return x.split(" = ")[1].replace("'", "")
    return None


def download_file(remote_path: str, use_cache: bool = False, local_name: Optional[str] = None) -> Optional[str]:
    """Download filename `remote_path` inside the mirror tree inside the `gef.config["gef.tempdir"]`.
    The tree architecture must be `gef.config["gef.tempdir"]/gef/<local_pid>/<remote_filepath>`.
    This allow a "chroot-like" tree format."""

    local_root = pathlib.Path(GLOBAL_TMP_DIR) / str(pid())
    if local_name is None:
        local_path = local_root / remote_path.strip(os.sep)
    else:
        local_path = local_root / local_name.strip(os.sep)

    if use_cache and local_path.exists():
        return str(local_path.absolute())

    try:
        local_path.parent.mkdir(parents=True, exist_ok=True)
        gdb.execute(f"remote get {remote_path} {local_path.absolute()}")
        local_path = str(local_path.absolute())
    except gdb.error:
        # fallback memory view
        with open(local_path, "w") as f:
            if is_32bit():
                f.write(f"00000000-ffffffff rwxp 00000000 00:00 0                    {get_filepath()}\n")
            else:
                f.write(f"0000000000000000-ffffffffffffffff rwxp 00000000 00:00 0                    {get_filepath()}\n")

    except Exception as e:
        err(f"download_file() failed: {e}")
        local_path = None

    return local_path


def is_alive() -> bool:
    """Check if GDB is running."""
    try:
        return gdb.selected_inferior().pid > 0
    except Exception:
        return False


def only_if_gdb_running(f: Callable) -> Callable:
    """Decorator wrapper to check if GDB is running."""

    @functools.wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if is_alive():
            return f(*args, **kwargs)
        else:
            warn("No debugging session active")

    return wrapper




import functools
import xmlrpc.client


def only_if_connected(f):
    @functools.wraps(f)
    def _only_if_connected(self, *args, **kwargs):
        if self.connected:
            return f(self, *args, **kwargs)

    return _only_if_connected


class DecompilerClient:
    def __init__(self, name="decompiler", host="localhost", port=3662, native_sym_support=True):
        self.name = name
        self.host = host
        self.port = port
        self.native_sym_support = native_sym_support
        self.server = None

    #
    # Server Ops
    #

    @property
    def connected(self):
        return True if self.server else False

    def connect(self, name=None, host=None, port=None) -> bool:
        """
        Connects to the remote decompiler.
        """
        self.name = name or self.name
        host = host or self.host
        port = port or self.port

        # create a decompiler server connection and test it
        retry = True
        try:
            self.server = xmlrpc.client.ServerProxy("http://{:s}:{:d}".format(host, port))
            self.server.ping()
            retry = False
        except:
            pass

        # the connection could fail because its a Ghidra connection on endpoint d2d
        if retry:
            try:
                self.server = xmlrpc.client.ServerProxy("http://{:s}:{:d}".format(host, port)).d2d
                self.server.ping()
            except (ConnectionRefusedError, AttributeError) as e:
                self.server = None
                # if we fail here, we fail overall
                return False

        self.decompiler_connected()
        return True

    def decompiler_connected(self):
        pass

    def decompiler_disconnected(self):
        pass

    #
    # Decompiler Interface
    #

    @only_if_connected
    def disconnect(self):
        try:
            self.server.disconnect()
        except Exception:
            pass

        self.server = None
        self.decompiler_disconnected()

    @only_if_connected
    def decompile(self, addr):
        return self.server.decompile(addr)

    @only_if_connected
    def function_data(self, addr):
        return self.server.function_data(addr)

    @property
    @only_if_connected
    def function_headers(self):
        return self.server.function_headers()

    @property
    @only_if_connected
    def global_vars(self):
        return self.server.global_vars()

    @property
    @only_if_connected
    def structs(self):
        return self.server.structs()

    @property
    @only_if_connected
    def breakpoints(self):
        return self.server.breakpoints()

    #
    # Client Setters
    #

    def update_global_vars(self):
        raise NotImplementedError

    def update_function_headers(self):
        raise NotImplementedError

    def update_function_data(self, addr):
        raise NotImplementedError



#
# ██████╗ ███████╗ ██████╗ ██████╗ ███╗   ███╗██████╗ ██████╗ ██████╗ ██████╗  ██████╗
# ██╔══██╗██╔════╝██╔════╝██╔═══██╗████╗ ████║██╔══██╗╚════██╗██╔══██╗██╔══██╗██╔════╝
# ██║  ██║█████╗  ██║     ██║   ██║██╔████╔██║██████╔╝ █████╔╝██║  ██║██████╔╝██║  ███╗
# ██║  ██║██╔══╝  ██║     ██║   ██║██║╚██╔╝██║██╔═══╝ ██╔═══╝ ██║  ██║██╔══██╗██║   ██║
# ██████╔╝███████╗╚██████╗╚██████╔╝██║ ╚═╝ ██║██║     ███████╗██████╔╝██████╔╝╚██████╔╝
# ╚═════╝ ╚══════╝ ╚═════╝ ╚═════╝ ╚═╝     ╚═╝╚═╝     ╚══════╝╚═════╝ ╚═════╝  ╚═════╝
#

import threading
import traceback

try:
    from PySide2.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QLineEdit, QMessageBox, \
        QGridLayout
    from PySide2.QtGui import Qt, QKeySequence

except ImportError:
    from PySide6.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QLineEdit, QMessageBox, \
        QGridLayout
    from PySide6.QtGui import Qt, QKeySequence


from binaryninjaui import (
    UIContext,
    UIAction,
    UIActionHandler,
    Menu,
)
from binaryninja import core_version

from .server import BinjaDecompilerServer

#
# UI
#

class ConfigDialog(QDialog):
    def __init__(self, bv, parent=None):
        super().__init__(parent)
        self.bv = bv
        self.setWindowTitle("Configure decomp2dbg")
        self._main_layout = QVBoxLayout()
        self._host_edit = None  # type:QLineEdit
        self._port_edit = None  # type:QLineEdit

        self._init_widgets()
        self.setLayout(self._main_layout)
        self.show()

    def _init_widgets(self):
        upper_layout = QGridLayout()

        host_label = QLabel(self)
        host_label.setText("Host")
        self._host_edit = QLineEdit(self)
        self._host_edit.setText("localhost")
        row = 0
        upper_layout.addWidget(host_label, row, 0)
        upper_layout.addWidget(self._host_edit, row, 1)
        row += 1

        port_label = QLabel(self)
        port_label.setText("Port")
        self._port_edit = QLineEdit(self)
        self._port_edit.setText("3662")
        upper_layout.addWidget(port_label, row, 0)
        upper_layout.addWidget(self._port_edit, row, 1)
        row += 1

        # buttons
        self._ok_button = QPushButton(self)
        self._ok_button.setText("OK")
        self._ok_button.setDefault(True)
        self._ok_button.clicked.connect(self._on_ok_clicked)
        cancel_button = QPushButton(self)
        cancel_button.setText("Cancel")
        cancel_button.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(cancel_button)

        # main layout
        self._main_layout.addLayout(upper_layout)
        self._main_layout.addLayout(buttons_layout)

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        host = self._host_edit.text()
        port = self._port_edit.text()

        if not host:
            QMessageBox(self).critical(None, "Invalid host",
                                       "Host cannot be empty."
                                       )
            return

        if not port:
            QMessageBox(self).critical(None, "Invalid port",
                                       "Port cannot be empty"
                                       )
            return

        decomp_server = BinjaDecompilerServer(self.bv)
        t = threading.Thread(target=decomp_server.start_xmlrpc_server, kwargs={'host': host, 'port': int(port)})
        t.daemon = True
        try:
            t.start()
        except Exception as e:
            QMessageBox(self).critical(None, "Error starting decomp2dbg Server", str(e))
            traceback.print_exc()
            return

        self.close()

    def _on_cancel_clicked(self):
        self.close()

#
# Plugin
#

class BinjaPlugin:
    def __init__(self):
        # controller stored by a binary view
        self._init_ui()

    def _init_ui(self):
        # config dialog
        configure_d2d_id = "decomp2dbg: Configure..."
        UIAction.registerAction(configure_d2d_id, QKeySequence(Qt.CTRL | Qt.SHIFT | Qt.Key_D))
        UIActionHandler.globalActions().bindAction(
            configure_d2d_id, UIAction(self._launch_config)
        )

        target_menu = "Tools" if int(core_version()[4:][:4]) < 3505 else "Plugins"
        Menu.mainMenu(target_menu).addAction(configure_d2d_id, "decomp2dbg")

    def _launch_config(self, bn_context):
        bv = bn_context.binaryView

        # configure
        dialog = ConfigDialog(bv)
        dialog.exec_()


BinjaPlugin()



from xmlrpc.server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler

from binaryninja import SymbolType, EntryRegisterValue
from binaryninja.binaryview import BinaryDataNotification
import binaryninja

#
# Binja Hooks
#


class DataNotification(BinaryDataNotification):
    def __init__(self, bv, server):
        super().__init__()
        self.bv = bv
        self.server = server  # type: BinjaDecompilerServer

    def symbol_updated(self, view, sym):
        if sym.type == SymbolType.FunctionSymbol:
            self.server.cache["function_headers"][str(sym.address)]["name"] = sym.name
        elif sym.type == SymbolType.DataSymbol:
            self.server.cache["global_vars"][str(sym.address)]["name"] = sym.name

#
# Server Code
#


class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/RPC2",)


class BinjaDecompilerServer:
    def __init__(self, bv, host=None, port=None):
        self.bv = bv
        self.host = host
        self.port = port

        # save the last line for bugging decomp mapping
        self._last_line = 0

        # cache changes so we don't need to regen content
        self.cache = {
            "global_vars": None,
            "function_headers": None
        }

        # make the server init cache data once
        self.function_headers()
        self.global_vars()

        # init hooks for cache
        notification = DataNotification(self.bv, self)
        self.bv.register_notification(notification)

    #
    # Public API
    #

    def decompile(self, addr: int):
        resp = {
            "decompilation": None,
            "curr_line": None,
            "func_name": None
        }
        funcs = self.bv.get_functions_containing(addr)
        if not funcs:
            return resp
        func = funcs[0]

        decomp = str(func.hlil).split("\n")
        if not decomp:
            return resp

        resp["decompilation"] = decomp
        resp["func_name"] = func.name

        # find the decompiled line closest to the current addr
        decomp_lines = func.get_low_level_il_at(addr).hlils
        if not decomp_lines:
            resp["curr_line"] = self._last_line
            return resp

        best_line = min(decomp_lines, key=lambda l: abs(l.address - addr))

        resp["curr_line"] = best_line.instr_index
        self._last_line = resp["curr_line"] if resp["curr_line"] != 0 else self._last_line
        return resp

    def function_data(self, addr: int):
        """
        Returns stack vars and func args

        """
        resp = {
            "reg_vars": {},
            "stack_vars": {}
        }

        funcs = self.bv.get_functions_containing(addr)
        if not funcs:
            return resp

        func = funcs[0]

        # get stack frame offset for x86
        frame_offset = 0
        if self.bv.arch.name == 'x86_64':
            frame_offset -= self.bv.arch.address_size
        elif self.bv.arch.name == 'x86':
            # handle inconsistent stack frame offsets
            current_frame = func.get_reg_value_at(addr, 'ebp')
            if current_frame.type != EntryRegisterValue.type:
                frame_offset = current_frame.value

        # get stack vars
        stack_vars = {}
        for stack_var in func.stack_layout:
            offset = frame_offset - stack_var.storage
            stack_vars[str(offset)] = {
                "name": stack_var.name,
                "type": str(stack_var.type)
            }

        # get reg vars
        reg_vars = {}
        for var in func.vars:
            if var.source_type != binaryninja.VariableSourceType.RegisterVariableSourceType or not var.name:
                continue

            reg_vars[var.name] = {
                "reg_name": self.bv.arch.get_reg_name(var.storage),
                "type": str(var.type)
            }

        resp["reg_vars"] = reg_vars
        resp["stack_vars"] = stack_vars

        return resp

    def function_headers(self):
        # check if a cache is available
        cache_headers = self.cache["function_headers"]
        if cache_headers:
            return cache_headers

        resp = {}
        for func in self.bv.functions:

            # Skip everything besides FunctionSymbol
            if func.symbol.type != SymbolType.FunctionSymbol:
                continue

            resp[str(func.start)] = {
                "name": func.name,
                "size": func.total_bytes
            }

        self.cache["function_headers"] = resp
        return resp

    def global_vars(self):
        # check if a cache is available
        cache_globals = self.cache["global_vars"]
        if cache_globals:
            return cache_globals

        resp = {}
        for addr, var in self.bv.data_vars.items():
            sym = self.bv.get_symbol_at(addr)
            name = sym.name if sym else "data_{:x}".format(addr)

            resp[str(addr)] = {
                "name": name
            }

        self.cache["global_vars"] = resp
        return resp

    def structs(self):
        resp = {}
        """
        # tuple of structure name and StructureType
        for t in self.bv.types.items():
            struct_name = t[0]
            struct = t[1]
            resp[struct_name] = {
                "size": struct.width
            }
            for member in struct.members:
                resp[struct_name][member.name] = {
                    "offset": member.offset,
                    "size": len(member)
                }
        """

        return resp

    def breakpoints(self):
        resp = {}
        return resp

    #
    # XMLRPC Server
    #

    def ping(self):
        return True

    def start_xmlrpc_server(self, host="localhost", port=3662):
        """
        Initialize the XMLRPC thread.
        """
        host = host or self.host
        port = port or self.port

        print("[+] Starting XMLRPC server: {}:{}".format(host, port))
        server = SimpleXMLRPCServer(
            (host, port),
            requestHandler=RequestHandler,
            logRequests=False,
            allow_none=True
        )
        server.register_introspection_functions()
        server.register_function(self.decompile)
        server.register_function(self.function_headers)
        server.register_function(self.function_data)
        server.register_function(self.global_vars)
        server.register_function(self.structs)
        server.register_function(self.breakpoints)
        server.register_function(self.ping)
        print("[+] Registered decompilation server!")
        while True:
            server.handle_request()



import sys

import pwndbg

from .gdb_client import GDBClient
from .decompiler_pane import DecompilerPane
from ...utils import *
import xmlrpc.client


class PwndbgDecompilerPane(DecompilerPane):
    def __init__(self, decompiler):
        super(PwndbgDecompilerPane, self).__init__(decompiler)

    def decompilation_text(self):
        """
        Display the current decompilation, with an arrow next to the current line.
        """
        output = []

        if not self.decompiler.connected:
            return output

        if not self.ready_to_display:
            err("Unable to decompile function")
            return output

        # configure based on source config
        past_lines_color = "gray"
        nb_lines = 6
        cur_line_color = "green"

        if len(self.decomp_lines) < nb_lines:
            nb_lines = len(self.decomp_lines)

        # use GEF source printing method
        for i in range(self.curr_line - nb_lines + 1, self.curr_line + nb_lines):
            if i < 0:
                continue

            if i < self.curr_line:
                output += ["{}".format(Color.colorify("  {:4d}\t {:s}".format(i + 1, self.decomp_lines[i], ), past_lines_color))]

            if i == self.curr_line:
                prefix = "{}{:4d}\t ".format(RIGHT_ARROW[1:], i + 1)
                output += [Color.colorify("{}{:s}".format(prefix, self.decomp_lines[i]), cur_line_color)]

            if i > self.curr_line:
                try:
                    output += ["  {:4d}\t {:s}".format(i + 1, self.decomp_lines[i], )]
                except IndexError:
                    break
        return output

    def context_gdecompiler(self, target=sys.stdout, with_banner=True, width=None):
        failed = True
        try:
            # triggers an update as well
            pane_title = self.title()
            failed = False
        except Exception as e:
            pane_title = f"decompiler: {e}"

        if pane_title is None:
            return []

        banner = [pwndbg.ui.banner(pane_title, target=target, width=width)] if with_banner else []
        if failed:
            return banner + ["decompilation error"]

        return banner + self.decompilation_text()


class PwndbgClient(GDBClient):
    def __init__(self):
        super(PwndbgClient, self).__init__()
        self.dec_pane = PwndbgDecompilerPane(self.dec_client)

        # if we are connected to ghidra
        if not isinstance(self.dec_client.server, xmlrpc.client.ServerProxy):
            # reset the type handlers pwndbg adds
            xmlrpc.client.Marshaller.dispatch[type(0)] = xmlrpc.client.Marshaller.dump_long

    def register_decompiler_context_pane(self, decompiler_name):
        pwndbg.commands.context.context_sections["g"] = self.dec_pane.context_gdecompiler
        pwndbg.commands.config_context_sections = pwndbg.lib.config.Parameter(
            f'context-sections',
            f'regs disasm code gdecompiler stack backtrace expressions',
            f'which context sections are displayed (controls order)'
        )

    def deregister_decompiler_context_pane(self, decompiler_name):
        del pwndbg.commands.context.context_sections["g"]
        pwndbg.commands.config_context_sections = pwndbg.lib.config.Parameter(
            f'context-sections',
            f'regs disasm code ghidra stack backtrace expressions',
            f'which context sections are displayed (controls order)'
        )



from xmlrpc.server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler


class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/RPC2",)


class DecompilerServer:
    def __init__(self, host=None, port=None):
        self.host = host
        self.port = port

    #
    # Public API
    #

    def decompile(self, addr: int):
        """
        Takes an addr which may be in a function. If addr is not in a function, a dict with the defined
        parameters below should be returned with None for each value. Decompilation should be the decompilation
        string of the function. curr_line should be the line number of that decompilation, starting at 0.

        Always returns a dict with the defined keys below, which may have None as their values.
        """
        resp = {
            "decompilation": str,
            "curr_line": int,
            "func_name": str
        }

        return resp

    def function_data(self, addr: int):
        """
        Returns stack vars and func args

        """
        resp = {
            "args": {
                0x0: {
                    "name": "example_name",
                    "type": "some_type",
                },
            },
            "stack_vars": {
                0x10: {
                    "name": "example_name",
                    "type": "some_type",
                },
            },
        }

        return resp

    def function_headers(self):
        resp = {
            0xdeadbeef: {
                "name": str,
                "size": int
            },
        }

        return resp

    def global_vars(self):
        resp = {
            0xdeadbeef: {
                "name": str
            },
        }

        return resp

    def structs(self):
        resp = {
            "example_struct_name": {
                "size": int,
                "example_member_name": {
                    "offset": int,
                    "size": int
                },
            },
        }

        return resp

    def breakpoints(self):
        resp = {
            0xdeadbeef: bool,
            0xdeadbeef+1: bool,
        }

        return resp

    #
    # XMLRPC Server
    #

    def ping(self):
        return True

    def start_xmlrpc_server(self, host="localhost", port=3662):
        """
        Initialize the XMLRPC thread.
        """
        host = host or self.host
        port = port or self.port

        print("[+] Starting XMLRPC server: {}:{}".format(host, port))
        server = SimpleXMLRPCServer(
            (host, port),
            requestHandler=RequestHandler,
            logRequests=False,
            allow_none=True
        )
        server.register_introspection_functions()
        server.register_function(self.decompile)
        server.register_function(self.function_headers)
        server.register_function(self.function_data)
        server.register_function(self.global_vars)
        server.register_function(self.structs)
        server.register_function(self.breakpoints)
        server.register_function(self.ping)
        print("[+] Registered decompilation server!")
        while True:
            server.handle_request()



# pylint: disable=missing-class-docstring
import os
import platform
import shutil
from pathlib import Path
import sys
from distutils.util import get_platform
from distutils.command.build import build as st_build

from setuptools import setup
from setuptools.command.develop import develop as st_develop


def _copy_decomp_plugins():
    local_plugins = Path("decompilers").absolute()
    decomp2dbg_loc = Path("decomp2dbg").absolute()
    pip_e_plugins = decomp2dbg_loc.joinpath("decompilers").absolute()

    local_d2d = Path("d2d.py").absolute()

    # clean the install location of symlink or folder
    shutil.rmtree(pip_e_plugins, ignore_errors=True)
    try:
        os.unlink(pip_e_plugins)
        os.unlink(decomp2dbg_loc.joinpath(local_d2d.name))
    except:
        pass

    # first attempt a symlink, if it works, exit early
    try:
        os.symlink(local_plugins, pip_e_plugins, target_is_directory=True)
        os.symlink(local_d2d, decomp2dbg_loc.joinpath(local_d2d.name))
        return
    except:
        pass

    # copy if symlinking is not available on target system
    try:
        shutil.copytree("decompilers", "decomp2dbg/decompilers")
        shutil.copy("d2d.py", "decomp2dbg/d2d.py")
    except:
        pass

class build(st_build):
    def run(self, *args):
        self.execute(_copy_decomp_plugins, (), msg="Copying binsync plugins")
        super().run(*args)

class develop(st_develop):
    def run(self, *args):
        self.execute(_copy_decomp_plugins, (), msg="Linking or copying local plugins folder")
        super().run(*args)


cmdclass = {
    "build": build,
    "develop": develop,
}

if 'bdist_wheel' in sys.argv and '--plat-name' not in sys.argv:
    sys.argv.append('--plat-name')
    name = get_platform()
    if 'linux' in name:
        sys.argv.append('manylinux2014_' + platform.machine())
    else:
        # https://www.python.org/dev/peps/pep-0425/
        sys.argv.append(name.replace('.', '_').replace('-', '_'))

setup(cmdclass=cmdclass)



from xmlrpc.server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
import os

import angr
from angr.analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator



class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/RPC2",)


class AngrDecompilerServer:
    def __init__(self, instance, host="localhost", port=3662):
        self._instance = instance
        self._workspace = self._instance.workspace
        self.host = host
        self.port = port

    #
    # Private
    #

    def rebase_addr(self, addr, down=False):
        rebased = addr
        base_addr = self._instance.project.loader.min_addr

        if down:
            rebased -= base_addr
        elif addr < base_addr:
            rebased += base_addr

        return rebased

    def _decompile_function(self, func):
        """
        Taken directly from BinSync
        """
        # check for known decompilation
        available = self._instance.kb.structured_code.available_flavors(func.addr)
        should_decompile = False
        if 'pseudocode' not in available:
            should_decompile = True
        else:
            cached = self._instance.kb.structured_code[(func.addr, 'pseudocode')]
            if isinstance(cached, DummyStructuredCodeGenerator):
                should_decompile = True

        if should_decompile:
            # recover direct pseudocode
            self._instance.project.analyses.Decompiler(func, flavor='pseudocode')

            # attempt to get source code if its available
            source_root = None
            if self._instance.original_binary_path:
                source_root = os.path.dirname(self._instance.original_binary_path)
            self._instance.project.analyses.ImportSourceCode(func, flavor='source', source_root=source_root)

        # grab newly cached pseudocode
        decomp = self._instance.kb.structured_code[(func.addr, 'pseudocode')].codegen
        return decomp

    #
    # Public API
    #

    def decompile(self, addr: int):
        """
        Takes an addr which may be in a function. If addr is not in a function, a dict with the defined
        parameters below should be returned with None for each value. Decompilation should be the decompilation
        string of the function. curr_line should be the line number of that decompilation, starting at 0.

        Always returns a dict with the defined keys below, which may have None as their values.
        """
        resp = {
            "decompilation": None
        }

        addr = self.rebase_addr(addr)
        func_addr = self._instance.cfg.get_any_node(addr, anyaddr=True).function_address
        func = self._instance.kb.functions[func_addr]
        decomp = self._decompile_function(func)
        pos = decomp.map_addr_to_pos.get_nearest_pos(addr)
        size = len(decomp.text)
        line_end = decomp.text.find("\n", pos)
        line_start = size - decomp.text[::-1].find("\n", size - line_end)
        decomp_lines = decomp.text.split('\n')
        for idx, line in enumerate(decomp_lines):
            if decomp.text[line_start:line_end] in line:
                break
        else:
            return resp

        resp["decompilation"] = decomp_lines
        resp["func_name"] = func.name
        resp["curr_line"] = idx
        return resp

    def function_data(self, addr: int):
        """
        Returns stack vars and func args

        """
        resp = {
            "reg_vars": {},
            "stack_vars": {}
        }

        addr = self.rebase_addr(addr)
        func_addr = self._instance.cfg.get_any_node(addr, anyaddr=True).function_address
        func = self._instance.kb.functions[func_addr]
        decomp = self._decompile_function(func)
        manager = decomp.cfunc.variable_manager
        for var in manager._unified_variables:
            if isinstance(var, angr.sim_variable.SimStackVariable):
                resp["stack_vars"][str(var.offset)] = {
                    "name": var.name,
                    "type": manager.get_variable_type(var).c_repr()
                }

        return resp

    def function_headers(self):
        resp = {}
        for addr, func in self._instance.kb.functions.items():
            resp[str(self.rebase_addr(addr, down=True))] = {
                "name": func.name,
                "size": func.size
            }

        return resp

    def global_vars(self):
        resp = {}

        return resp

    def structs(self):
        resp = {}

        return resp

    def breakpoints(self):
        resp = {}

        return resp

    #
    # XMLRPC Server
    #

    def ping(self):
        return True

    def start_xmlrpc_server(self, host="localhost", port=3662):
        """
        Initialize the XMLRPC thread.
        """
        host = host or self.host
        port = port or self.port

        print("[+] Starting XMLRPC server: {}:{}".format(host, port))
        server = SimpleXMLRPCServer(
            (host, port),
            requestHandler=RequestHandler,
            logRequests=False,
            allow_none=True
        )
        server.register_introspection_functions()
        server.register_function(self.decompile)
        server.register_function(self.function_headers)
        server.register_function(self.function_data)
        server.register_function(self.global_vars)
        server.register_function(self.structs)
        server.register_function(self.breakpoints)
        server.register_function(self.ping)
        print("[+] Registered decompilation server!")
        while True:
            server.handle_request()



from .gdb_client import GDBClient
from .utils import is_remote_debug, checksec


class GEFClient(GDBClient):
    def __init__(self, ctx_pane_registrar, gef_print, gef_ref): #gef_config, gef_memory):
        super(GEFClient, self).__init__()
        self.ctx_pane_registrar = ctx_pane_registrar
        self.dec_pane.print = gef_print
        self.gef_config = gef_ref.config
        self.gef_ref = gef_ref

    def register_decompiler_context_pane(self, decompiler_name):
        self.ctx_pane_registrar("decompilation", self.dec_pane.display_pane, self.dec_pane.title)

    def deregister_decompiler_context_pane(self, decompiler_name):
        self.gef_config["context.layout"] = self.gef_config["context.layout"].replace(" decompilation", "")

    def find_text_segment_base_addr(self, is_remote=False):
        if is_remote:
            elf_file = str(self.gef_ref.session.remote.lfile)
            elf_virtual_path = str(self.gef_ref.session.remote.file)
        else:
            elf_file = str(self.gef_ref.session.file)
            elf_virtual_path = str(self.gef_ref.session.file)

        vmmap = self.gef_ref.memory.maps
        base_address = min(x.page_start for x in vmmap if x.path == elf_virtual_path)

        return base_address
    
    @property
    def is_pie(self):
        elf_file = self.gef_ref.session.remote.lfile if is_remote_debug() else self.gef_ref.session.file
        checksec_status = checksec(str(elf_file))
        return checksec_status["PIE"]  # if pie we will have offset instead of abs address.



import textwrap
import argparse
import contextlib
import io

import gdb

from ..client import DecompilerClient
from ...utils import *
from .utils import *
from .symbol_mapper import SymbolMapper
from .decompiler_pane import DecompilerPane

#
# Decompiler Client Interface
#


class GDBDecompilerClient(DecompilerClient):
    def __init__(self, gdb_client, name="decompiler", host="127.0.0.1", port=3662):
        super(GDBDecompilerClient, self).__init__(name=name, host=host, port=port)
        self.gdb_client: "GDBClient" = gdb_client
        self.symbol_mapper = SymbolMapper()
        self._is_pie = None
        self._lvar_bptr = None

    @property
    @lru_cache()
    def text_base_addr(self):
        return self.gdb_client.base_addr_start

    @property
    def is_pie(self):
        if self._is_pie is None:
            self._is_pie = self.gdb_client.is_pie

        return self._is_pie

    def rebase_addr(self, addr, up=False):
        corrected_addr = addr
        if self.is_pie or self.gdb_client.base_manually_set:
            if up:
                corrected_addr += self.text_base_addr
            else:
                corrected_addr -= self.text_base_addr

        return corrected_addr

    def decompiler_connected(self):
        self.gdb_client.on_decompiler_connected(self.name)

    def decompiler_disconnected(self):
        self.gdb_client.on_decompiler_disconnected(self.name)

    def update_symbols(self):
        self.symbol_mapper.text_base_addr = self.text_base_addr

        global_vars, func_headers = self.update_global_vars(), self.update_function_headers()
        syms_to_add = []
        sym_name_set = set()
        global_var_size = 8

        if not self.native_sym_support:
            err("Native symbol support is required to run decomp2dbg, assure you have coreutils installed.")
            return False

        # add symbols with native support if possible
        for addr, func in func_headers.items():
            syms_to_add.append((func["name"], int(addr, 0), "function", func["size"]))
            sym_name_set.add(func["name"])

        for addr, global_var in global_vars.items():
            clean_name = re.sub(r"[^a-zA-Z0-9_]", "_", global_var['name'])
            # never re-add globals with the same name as a func
            if clean_name in sym_name_set:
                continue

            syms_to_add.append((clean_name, int(addr, 0), "object", global_var_size))

        try:
            self.symbol_mapper.add_native_symbols(syms_to_add)
        except Exception as e:
            err(f"Failed to set symbols natively: {e}")
            self.native_sym_support = False
            return False

        return True

    def update_global_vars(self):
        return self.global_vars

    def update_function_headers(self):
        return self.function_headers

    def _clean_type_str(self, type_str):
        if "__" in type_str:
            type_str = type_str.replace("__", "")
            idx = type_str.find("[")
            if idx != -1:
                type_str = type_str[:idx] + "_t" + type_str[idx:]
            else:
                type_str += "_t"
        type_str = type_str.replace("unsigned ", "u")

        return type_str

    def _find_local_var_base_ptr(self):
        if self._lvar_bptr is not None:
            return self._lvar_bptr

        candidates = ["$rbp", "$ebp", "$bp", "$fp"]
        for candidate in candidates:
            # the register must exist
            try:
                val = gdb.parse_and_eval(candidate)
            except Exception:
                continue

            # the register must have a real value
            try:
                int(str(val), 16)
            except Exception:
                continue

            self._lvar_bptr = candidate
            return self._lvar_bptr


    def update_function_data(self, addr):
        func_data = self.function_data(addr)
        reg_vars = func_data.get("reg_vars", {})
        stack_vars = func_data.get("stack_vars", {})

        for name, var in reg_vars.items():
            type_str = self._clean_type_str(var['type'])
            reg_name = var['reg_name']
            expr = f"""(({type_str}) (${reg_name})"""

            try:
                val = gdb.parse_and_eval(expr)
                gdb.execute(f"set ${name} = {val}")
                type_unknown = False
            except Exception:
                type_unknown = True

            if type_unknown:
                try:
                    gdb.execute(f"set ${name} = (${reg_name})")
                except Exception:
                    continue

        for offset, stack_var in stack_vars.items():
            offset = abs(int(offset, 0))
            type_str = self._clean_type_str(stack_var['type'])
            lvar_bptr = self._find_local_var_base_ptr()
            if lvar_bptr == "$rbp":
                offset -= 8
            elif lvar_bptr == "$ebp":
                offset -= 4

            expr = f"""({type_str}*) ({lvar_bptr} - {offset})"""
            var_name = stack_var['name']

            try:
                gdb.execute(f"set ${var_name} = " + expr)
                type_unknown = False
            except Exception:
                type_unknown = True

            if type_unknown:
                try:
                    gdb.execute(f"set ${var_name} = ($fp - {offset})")
                except Exception:
                    continue


#
# Command Interface
#

class DecompilerCommand(gdb.Command):
    def __init__(self, decompiler, gdb_client):
        super(DecompilerCommand, self).__init__("decompiler", gdb.COMMAND_USER)
        self.decompiler = decompiler
        self.gdb_client = gdb_client
        self.arg_parser = self._init_arg_parser()

    @only_if_gdb_running
    def invoke(self, arg, from_tty):
        raw_args = arg.split()
        try:
            f = io.StringIO()
            with contextlib.redirect_stderr(f):
                args = self.arg_parser.parse_args(raw_args)
        except (RuntimeError, RuntimeWarning):
            return
        except Exception as e:
            err(f"Error parsing args: {e}")
            self.arg_parser.print_help()
            return

        self._handle_cmd(args)

    @staticmethod
    def _init_arg_parser():
        parser = argparse.ArgumentParser()
        commands = ["connect", "disconnect", "info"]
        parser.add_argument(
            'command', type=str, choices=commands, help="""
            Commands:
            [connect]: connects a decompiler by name, with optional host, port, and base address.
            [disconnect]: disconnects a decompiler by name, destroyed decompilation panel.
            [info]: lists useful info about connected decompilers
            """
        )
        parser.add_argument(
            'decompiler_name', type=str, nargs="?", help="""
            The name of the decompiler, which can be anything you like. It's suggested
            to use sematic and numeric names like: 'ida2' or 'ghidra1'. Optional when doing 
            the info command.
            """
        )
        parser.add_argument(
            '--host', type=str, default="localhost"
        )
        parser.add_argument(
            '--port', type=int, default=3662
        )
        parser.add_argument(
            '--base-addr-start', type=lambda x: int(x,0)
        )
        parser.add_argument(
            '--base-addr-end', type=lambda x: int(x,0)
        )

        return parser

    def _handle_cmd(self, args):
        cmd = args.command
        handler_str = f"_handle_{cmd}"
        handler = getattr(self, handler_str)
        handler(args)

    def _handle_connect(self, args):
        if not args.decompiler_name:
            err("You must provide a decompiler name when using this command!")
            return

        if (args.base_addr_start and not args.base_addr_end) or (args.base_addr_end and not args.base_addr_start):
            err("You must use --base-addr-start and --base-addr-end together")
            return
        elif args.base_addr_start and args.base_addr_end:
            if args.base_addr_start > args.base_addr_end:
                err("Your provided base-addr-start must be smaller than your base-addr-end")
                return

            self.gdb_client.base_addr_start = args.base_addr_start
            self.gdb_client.base_addr_end = args.base_addr_end
            self.gdb_client.base_manually_set = True

        self.gdb_client.name = args.decompiler_name
        connected = self.decompiler.connect(name=args.decompiler_name, host=args.host, port=args.port)
        if not connected:
            err("Failed to connect to decompiler!")
            return

        info("Connected to decompiler!")

    def _handle_disconnect(self, args):
        if not args.decompiler_name:
            err("You must provide a decompiler name when using this command!")
            return

        self.decompiler.disconnect()
        info("Disconnected decompiler!")

    def _handle_info(self, args):
        info("Decompiler Info:")
        print(textwrap.dedent(
            f"""\
            Name: {self.gdb_client.name}
            Base Addr Start: {hex(self.gdb_client.base_addr_start) 
            if isinstance(self.gdb_client.base_addr_start, int) else self.gdb_client.base_addr_start}
            Base Addr End: {hex(self.gdb_client.base_addr_end)
            if isinstance(self.gdb_client.base_addr_end, int) else self.gdb_client.base_addr_end}
            """
        ))


class GDBClient:
    def __init__(self):
        self.dec_client = GDBDecompilerClient(self)
        self.cmd_interface = DecompilerCommand(self.dec_client, self)
        self.dec_pane = DecompilerPane(self.dec_client)

        self.name = None
        self.base_manually_set = False
        self.base_addr_start = None
        self.base_addr_end = None

    def __del__(self):
        del self.cmd_interface

    def register_decompiler_context_pane(self, decompiler_name):
        gdb.events.stop.connect(self.dec_pane.display_pane_and_title)

    def deregister_decompiler_context_pane(self, decompiler_name):
        gdb.events.stop.disconnect(self.dec_pane.display_pane_and_title)

    def find_text_segment_base_addr(self, is_remote=False):
        return find_text_segment_base_addr(is_remote=is_remote)

    @property
    def is_pie(self):
        checksec_status = checksec(get_filepath())
        return checksec_status["PIE"]  # if pie we will have offset instead of abs address.

    #
    # Event Handlers
    #

    def on_decompiler_connected(self, decompiler_name):
        if self.base_addr_start is None:
            self.base_addr_start = self.find_text_segment_base_addr(is_remote=is_remote_debug())
        self.dec_client.update_symbols()
        self.register_decompiler_context_pane(decompiler_name)

    def on_decompiler_disconnected(self, decompiler_name):
        self.deregister_decompiler_context_pane(decompiler_name)
        self.name = None
        self.base_addr_start = None
        self.base_addr_end = None



from ...utils import *
from .utils import *
import tempfile
from elftools.elf.elffile import ELFFile
import gdb


class SymbolMapper:
    """
    A binary search dict implementation for ranges. Symbols will map for a range and we need to
    be able to lookup addresses in the middle of the range fast
    """

    __slots__ = (
        'text_base_addr',
        '_elf_cache',
        '_objcopy',
        '_gcc',
        '_last_sym_files',
        '_sym_file_ctr'
    )

    def __init__(self):
        self.text_base_addr = None
        self._elf_cache = {}
        self._objcopy = None
        self._gcc = None
        self._last_sym_files = set()
        self._sym_file_ctr = 0


    #
    # Native Symbol Support (Linux Only)
    # Inspired by Bata24
    #

    def add_native_symbols(self, sym_info_list):
        """
        Adds a list of symbols to gdb's internal symbol listing. Only function and global symbols are supported.
        Symbol info looks like:
        (symbol_name: str, base_addr: int, sym_type: str, size: int)
        If you don't know the size, pass 0.

        Explanation of how this works:
        Adding symbols to GDB is non-trivial, it requires the use of an entire object file. Because of its
        difficulty, this is currently only supported on ELFs. When adding a symbol, we use two binutils,
        gcc and objcopy. After making a small ELF, we strip it of everything but needed sections. We then
        use objcopy to one-by-one add a symbol to the file. Objcopy does not support sizing, so we do a byte
        patch on the binary to allow for a real size. Finally, the whole object is read in with the default
        gdb command: add-symbol-file.
        """

        if not self.check_native_symbol_support():
            err("Native symbol support not supported on this platform.")
            info("If you are on Linux and want native symbol support make sure you have gcc and objcopy.")
            return False

        if self.text_base_addr is None:
            err("Base address of the binary has not been discovered yet, please run the binary and try again.")
            return False

        # info("{:d} symbols will be added".format(len(sym_info_list)))
        self._delete_old_sym_files()

        # add each symbol into a mass symbol commit
        max_commit_size = 1500
        supported_types = ["function", "object"]

        objcopy_cmds = []
        queued_sym_sizes = {}
        fname = self._construct_small_elf()
        for i, (name, addr, typ, size) in enumerate(sym_info_list):
            if typ not in supported_types:
                warn(f"Skipping symbol {name}, type is not supported: {typ}")
                continue

            # queue the sym for later use
            queued_sym_sizes[i % max_commit_size] = size

            # absolute addressing
            #if addr >= self.text_base_addr:
            #    addr_str = "{:#x}".format(addr)
            # relative addressing
            #else:

            # you always want relative adressing
            addr_str = ".text:{:#x}".format(addr)

            # create a symbol command for the symbol
            objcopy_cmds.append(
                "--add-symbol '{name}'={addr_str},global,{type_flag}".format(
                    name=name, addr_str=addr_str, type_flag=typ
                )
            )

            # batch commit
            if i > 1 and i % max_commit_size == 0:
                # add the queued symbols
                self._add_symbol_file(fname, objcopy_cmds, self.text_base_addr, queued_sym_sizes)

                # re-init queues and elf
                fname = self._construct_small_elf()
                objcopy_cmds = []
                queued_sym_sizes = {}

        # commit remaining symbol commands
        if objcopy_cmds:
            self._add_symbol_file(fname, objcopy_cmds, self.text_base_addr, queued_sym_sizes)

        return True

    def check_native_symbol_support(self):
        # validate binutils bins exist
        try:
            self._gcc = which("gcc")
            self._objcopy = which("objcopy")
        except FileNotFoundError as e:
            err(f"Binutils binaries not found: {e}")
            return False

        return True

    def _delete_old_sym_files(self):
        for sym_file in self._last_sym_files:
            try:
                gdb.execute(f"remove-symbol-file {sym_file}")
            except Exception as e:
                pass

        self._last_sym_files = set()
        self._sym_file_ctr = 0

    def _construct_small_elf(self):
        if self._elf_cache:
            new_name = self._elf_cache["fname"]+str(self._sym_file_ctr)
            open(new_name, "wb").write(self._elf_cache["data"])
            self._sym_file_ctr += 1
            self._last_sym_files.add(new_name)
            return new_name

        # compile a small elf for symbol loading
        fd, fname = tempfile.mkstemp(dir="/tmp", suffix=".c")
        os.fdopen(fd, "w").write("int main() {}")
        # os.system(f"{self._gcc} {fname} -no-pie -o {fname}.debug")
        os.system(f"{self._gcc} {fname} -o {fname}.debug")
        # destroy the source file
        os.unlink(f"{fname}")

        # delete unneeded sections from object file
        os.system(f"{self._objcopy} --only-keep-debug {fname}.debug")
        os.system(f"{self._objcopy} --strip-all {fname}.debug")

        elf = ELFFile(open(f'{fname}.debug', 'rb'))

        required_sections = [".text", ".interp", ".rela.dyn", ".dynamic", ".bss"]
        for s in elf.iter_sections():
            # keep some required sections
            if s.name in required_sections:
                continue
            os.system(f"{self._objcopy} --remove-section={s.name} {fname}.debug 2>/dev/null")

        # cache the small object file for use
        self._elf_cache["fname"] = fname + ".debug"

        # add it to known sym files
        self._last_sym_files.add(self._elf_cache["fname"])

        self._elf_cache["data"] = open(self._elf_cache["fname"], "rb").read()
        return self._elf_cache["fname"]

    def _force_update_text_size(self, stream, elf, elf_data, new_size):
        text_sect = elf.get_section_by_name('.text')

        for count in range(elf['e_shnum']):
            sect_off = elf['e_shoff'] + count * elf['e_shentsize']
            stream.seek(sect_off)
            section_header = elf.structs.Elf_Shdr.parse_stream(stream)
            if section_header['sh_name'] == text_sect['sh_name']:
                break

        patch = struct.pack("<Q", new_size)
        elf_data[sect_off+32: sect_off + 32 + len(patch)] = patch
        return elf_data

    def _force_update_sym_sizes(self, fname, queued_sym_sizes):
        # use pyelftools to obtain accurate offsets and assign symbol sizes

        stream = open(fname, 'rb')
        elf_data = bytearray(stream.read())
        stream.close()

        stream = open(fname, 'rb')
        elf = ELFFile(stream)

        # patch .text to seem large enough for any function
        elf_data = self._force_update_text_size(stream, elf, elf_data, 0xFFFFFF)

        # find the symbol table
        section = elf.get_section_by_name('.symtab')
        if not section:
            return

            # locate the location of the symbols size in the symtab
        tab_offset = section['sh_offset']

        # see if symbols already exist to skip
        for skip_off, sym in enumerate(section.iter_symbols()):
            if sym.name:
                break
        else:
            # this should never happen
            skip_off = 1
        skip_off -= 1

        # NOTE: 64-bit elf checks are redundant as of now because all debug files
        # generated are 64-bit regardless of the original binary. Does not seem to
        # have any big problems as of now. May consider removing it or generating 32-bit
        # debug binaries if necessary for this code to remain relevant.
        sym_data_size = 24 if elf.elfclass == 64 else 16
        sym_size_off = sym_data_size - 8

        for i, size in queued_sym_sizes.items():
            # skip sizes of 0
            if not size:
                continue

            # compute offset
            sym_size_loc = tab_offset + sym_data_size * (i + 1 + skip_off) + sym_size_off
            pack_str = "<Q" if elf.elfclass == 64 else "<I"
            # write the new size
            updated_size = struct.pack(pack_str, size)
            elf_data[sym_size_loc:sym_size_loc + len(updated_size)] = updated_size

        # write data back to elf
        open(fname, "wb").write(elf_data)

    def _add_symbol_file(self, fname, cmd_string_arr, text_base, queued_sym_sizes):
        # add the symbols through copying
        cmd_string = ' '.join(cmd_string_arr)
        os.system(f"{self._objcopy} {cmd_string} {fname}")

        # force update the size of each symbol
        self._force_update_sym_sizes(fname, queued_sym_sizes)

        gdb.execute(f"add-symbol-file {fname} {text_base:#x}", to_string=True)

        os.unlink(fname)
        return




import textwrap
from urllib.request import urlretrieve

import pkg_resources
from pathlib import Path

from binsync.installer import Installer


class Decomp2dbgInstaller(Installer):
    def __init__(self):
        super(Decomp2dbgInstaller, self).__init__(targets=Installer.DECOMPILERS + ('gdb',))
        self.plugins_path = Path(
            pkg_resources.resource_filename("decomp2dbg", f"decompilers")
        )
    
    def display_prologue(self):
        print(textwrap.dedent("""
               __                               ___       ____         
          ____/ /__  _________  ____ ___  ____ |__ \ ____/ / /_  ____ _
         / __  / _ \/ ___/ __ \/ __ `__ \/ __ \__/ // __  / __ \/ __ `/
        / /_/ /  __/ /__/ /_/ / / / / / / /_/ / __// /_/ / /_/ / /_/ / 
        \__,_/\___/\___/\____/_/ /_/ /_/ .___/____/\__,_/_.___/\__, /  
                                      /_/                     /____/   
        Now installing decomp2dbg...
        Please input decompiler/debugger install paths as prompted. Enter nothing to either use
        the default install path if one exist, or to skip.
        """))

    def install_gdb(self, path=None):
        path = super().install_gdb(path=None)
        if path is None:
            return None

        d2d_script_path_pkg = self.plugins_path.parent.joinpath("d2d.py")
        with open(path, "r") as fp:
            init_contents = fp.read()
            
        write_str = f"source {str(d2d_script_path_pkg.absolute())}"
        if write_str in init_contents:
            self.warn("gdbinit already contains d2d source...")
            return None

        with open(path, "a") as fp:
            fp.write(f"\n{write_str}\n")

        return path

    def install_ida(self, path=None):
        ida_plugin_path = super().install_ida(path=path)
        if ida_plugin_path is None:
            return

        src_d2d_ida_pkg = self.plugins_path.joinpath("d2d_ida").joinpath("d2d_ida")
        src_d2d_ida_py = self.plugins_path.joinpath("d2d_ida").joinpath("d2d_ida.py")
        dst_d2d_ida_pkg = ida_plugin_path.joinpath("d2d_ida")
        dst_d2d_ida_py = ida_plugin_path.joinpath("d2d_ida.py")
        self.link_or_copy(src_d2d_ida_pkg, dst_d2d_ida_pkg, is_dir=True)
        self.link_or_copy(src_d2d_ida_py, dst_d2d_ida_py)
        return dst_d2d_ida_pkg

    def install_angr(self, path=None):
        angr_plugin_path = super().install_angr(path=path)
        if angr_plugin_path is None:
            return None

        src_d2d_angr_pkg = self.plugins_path.joinpath("d2d_angr")
        dst_d2d_angr_pkg = angr_plugin_path.joinpath("d2d_angr")
        self.link_or_copy(src_d2d_angr_pkg, dst_d2d_angr_pkg, is_dir=True)
        return dst_d2d_angr_pkg

    def install_ghidra(self, path=None):
        ghidra_path = super().install_ghidra(path=path)
        if ghidra_path is None:
            return None

        download_url = "https://github.com/mahaloz/decomp2dbg/releases/latest/download/d2d-ghidra-plugin.zip"
        dst_path = ghidra_path.joinpath("d2d-ghidra-plugin.zip")
        urlretrieve(download_url, dst_path)
        return dst_path

    def install_binja(self, path=None):
        binja_plugin_path = super().install_binja(path=path)
        if binja_plugin_path is None:
            return None

        src_path = self.plugins_path.joinpath("d2d_binja")
        dst_path = binja_plugin_path.joinpath("d2d_binja")
        self.link_or_copy(src_path, dst_path, is_dir=True)
        return dst_path



#
# ██████╗ ███████╗ ██████╗ ██████╗ ███╗   ███╗██████╗ ██████╗  ██████╗ ███████╗███████╗
# ██╔══██╗██╔════╝██╔════╝██╔═══██╗████╗ ████║██╔══██╗╚════██╗██╔════╝ ██╔════╝██╔════╝
# ██║  ██║█████╗  ██║     ██║   ██║██╔████╔██║██████╔╝ █████╔╝██║  ███╗█████╗  █████╗
# ██║  ██║██╔══╝  ██║     ██║   ██║██║╚██╔╝██║██╔═══╝ ██╔═══╝ ██║   ██║██╔══╝  ██╔══╝
# ██████╔╝███████╗╚██████╗╚██████╔╝██║ ╚═╝ ██║██║     ███████╗╚██████╔╝███████╗██║
# ╚═════╝ ╚══════╝ ╚═════╝ ╚═════╝ ╚═╝     ╚═╝╚═╝     ╚══════╝ ╚═════╝ ╚══════╝╚═╝
#                                IDA Server
#
# by mahaloz, 2021.
#
from xmlrpc.server import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler
import threading
import functools

import ida_hexrays, ida_funcs, idc, ida_pro, ida_lines, idaapi, idautils, ida_segment


class RequestHandler(SimpleXMLRPCRequestHandler):
    rpc_paths = ("/RPC2",)
#
# Wrappers for IDA Main thread r/w operations
#
#
# a special note about these functions:
# Any operation that needs to do some type of write to the ida db (idb), needs to be in the main
# thread due to some ida constraints. Sometimes reads also need to be in the main thread.
#


def is_mainthread():
    """
    Return a bool that indicates if this is the main application thread.
    """
    return isinstance(threading.current_thread(), threading._MainThread)


def execute_sync(func, sync_type):
    """
    Synchronize with the disassembler for safe database access.
    Modified from https://github.com/vrtadmin/FIRST-plugin-ida
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> object:
        output = [None]

        #
        # this inline function definition is technically what will execute
        # in the context of the main thread. we use this thunk to capture
        # any output the function may want to return to the user.
        #

        def thunk():
            output[0] = func(*args, **kwargs)
            return 1

        if is_mainthread():
            thunk()
        else:
            idaapi.execute_sync(thunk, sync_type)

        # return the output of the synchronized execution
        return output[0]
    return wrapper


def execute_read(func):
    return execute_sync(func, idaapi.MFF_READ)


def execute_write(func):
    return execute_sync(func, idaapi.MFF_WRITE)


def execute_ui(func):
    return execute_sync(func, idaapi.MFF_FAST)


#
# Decompilation API
#

class IDADecompilerServer:
    def __init__(self, host=None, port=None):
        self.host = host
        self.port = port
        self.cache = {
            "global_vars": None,
            "function_headers": None
        }
        self._base_addr = None

        # make the server init cache data once
        self.function_headers()
        self.global_vars()

    def rebase_addr(self, addr, down=False):
        if self._base_addr is None:
            self._base_addr = idaapi.get_imagebase()

        rebased = addr
        if down:
            rebased -= self._base_addr
        elif addr < self._base_addr:
            rebased += self._base_addr

        return rebased


    @execute_read
    def decompile(self, addr):
        addr = self.rebase_addr(addr)
        resp = {
            "decompilation": None,
            "curr_line": None,
            "func_name": None,
        }

        # get the function
        ida_func = ida_funcs.get_func(addr)
        if not ida_func:
            return resp

        func_addr = ida_func.start_ea
        func_name = idc.get_func_name(func_addr)

        # attempt decompilation
        try:
            cfunc = ida_hexrays.decompile(func_addr)
        except Exception as e:
            return resp

        # locate decompilation line
        item = cfunc.body.find_closest_addr(addr)
        y_holder = idaapi.int_pointer()
        if not cfunc.find_item_coords(item, None, y_holder):
            # if idaapi failes, try a dirty search!
            up = cfunc.body.find_closest_addr(addr - 0x10)
            if not cfunc.find_item_coords(up, None, y_holder):
                down = cfunc.body.find_closest_addr(addr + 0x10)
                if not cfunc.find_item_coords(down, None, y_holder):
                    return resp

        cur_line_num = y_holder.value()

        # decode ida lines
        enc_lines = cfunc.get_pseudocode()
        decomp_lines = [idaapi.tag_remove(l.line) for l in enc_lines]

        resp["decompilation"] = decomp_lines
        resp["curr_line"] = cur_line_num
        resp["func_name"] = func_name

        return resp

    @execute_read
    def function_data(self, addr):
        addr = self.rebase_addr(addr)
        resp = {
            "stack_vars": None,
            "reg_vars": None
        }

        # get the function
        ida_func = ida_funcs.get_func(addr)
        if not ida_func:
            return resp
        func_addr = ida_func.start_ea

        # attempt decompilation
        try:
            cfunc = ida_hexrays.decompile(func_addr)
        except Exception as e:
            return resp

        # get var info
        stack_vars = {}
        reg_vars = {}

        for var in cfunc.lvars:
            if not var.name:
                continue

            # stack variables
            if var.is_stk_var():
                offset = cfunc.mba.stacksize - var.location.stkoff()
                stack_vars[str(offset)] = {
                    "name": var.name,
                    "type": str(var.type())
                }

            # register variables
            elif var.is_reg_var():
                regnum = var.get_reg1()
                reg_name = idaapi.get_mreg_name(regnum, var.width)

                reg_vars[var.name] = {
                    "reg_name": reg_name,
                    "type": str(var.type())
                }
                pass

        resp["stack_vars"] = stack_vars
        resp["reg_vars"] = reg_vars

        return resp

    @execute_read
    def function_headers(self):
        # check if a cache is available
        cache_headers = self.cache["function_headers"]
        if cache_headers:
            return cache_headers

        resp = {}
        # no cache, compute it
        for f_addr in idautils.Functions():
            # assure the function is not a linked library function
            if not ((idc.get_func_flags(f_addr) & idc.FUNC_LIB) == idc.FUNC_LIB):
                func_name = ida_funcs.get_func_name(f_addr)
                if not isinstance(func_name, str):
                    continue

                # double check for .plt or .got names
                if func_name.startswith(".") or "@" in func_name:
                    continue

                func_size = ida_funcs.get_func(f_addr).size()
                resp[str(self.rebase_addr(f_addr, down=True))] = {
                    "name": func_name,
                    "size": func_size
                }

        self.cache["function_headers"] = resp
        return resp

    @execute_read
    def global_vars(self):
        # check if a cache is available
        cache_globals = self.cache["global_vars"]
        if cache_globals:
            return cache_globals

        resp = {}
        known_segs = [".data", ".bss"]
        for seg_name in known_segs:
            seg = idaapi.get_segm_by_name(seg_name)
            if not seg:
                continue

            for seg_ea in range(seg.start_ea, seg.end_ea):
                xrefs = idautils.XrefsTo(seg_ea)
                try:
                    next(xrefs)
                except StopIteration:
                    continue

                name = idaapi.get_name(seg_ea)
                if not name:
                    continue

                resp[str(self.rebase_addr(seg_ea, down=True))] = {
                    "name": name
                }

        self.cache["global_vars"] = resp
        return resp

    @execute_read
    def structs(self):
        resp = {}
        for i in range(idaapi.get_struc_qty()):
            struct_id = idaapi.get_struc_by_idx(i)
            struct = idaapi.get_struc(struct_id)
            struct_info = {
                "name": idaapi.get_struc_name(struct.id),
                "members": []
            }

            for member in struct.members:
                member_info = {
                    "name": idaapi.get_member_name(member.id)
                }

                tif = idaapi.tinfo_t()
                if idaapi.get_member_tinfo(tif, member):
                    member_info["type"] = tif.__str__()
                member_info["size"] = tif.get_size()
                struct_info["members"].append(member_info)

            resp["struct_info"].append(struct_info)
        return resp

    def breakpoints(self):
        resp = {}
        return resp

    #
    # XMLRPC Server
    #

    def ping(self):
        return True

    def start_xmlrpc_server(self, host="localhost", port=3662):
        """
        Initialize the XMLRPC thread.
        """
        host = host or self.host
        port = port or self.port

        print("[+] Starting XMLRPC server: {}:{}".format(host, port))
        server = SimpleXMLRPCServer(
            (host, port),
            requestHandler=RequestHandler,
            logRequests=False,
            allow_none=True
        )
        server.register_introspection_functions()
        server.register_function(self.decompile)
        server.register_function(self.function_headers)
        server.register_function(self.function_data)
        server.register_function(self.global_vars)
        server.register_function(self.structs)
        server.register_function(self.breakpoints)
        server.register_function(self.ping)
        print("[+] Registered decompilation server!")
        while True:
            server.handle_request()



#
# ██████╗ ███████╗ ██████╗ ██████╗ ███╗   ███╗██████╗ ██████╗ ██████╗ ██████╗  ██████╗
# ██╔══██╗██╔════╝██╔════╝██╔═══██╗████╗ ████║██╔══██╗╚════██╗██╔══██╗██╔══██╗██╔════╝
# ██║  ██║█████╗  ██║     ██║   ██║██╔████╔██║██████╔╝ █████╔╝██║  ██║██████╔╝██║  ███╗
# ██║  ██║██╔══╝  ██║     ██║   ██║██║╚██╔╝██║██╔═══╝ ██╔═══╝ ██║  ██║██╔══██╗██║   ██║
# ██████╔╝███████╗╚██████╗╚██████╔╝██║ ╚═╝ ██║██║     ███████╗██████╔╝██████╔╝╚██████╔╝
# ╚═════╝ ╚══════╝ ╚═════╝ ╚═════╝ ╚═╝     ╚═╝╚═╝     ╚══════╝╚═════╝ ╚═════╝  ╚═════╝
#

import traceback
import threading

from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QLineEdit, QMessageBox, QGridLayout

import idaapi, ida_idp, ida_struct, idc, ida_enum

from .server import IDADecompilerServer

decomp_server: IDADecompilerServer = None


#
# Update Hooks
#

class IDBHooks(ida_idp.IDB_Hooks):
    def __init__(self):
        ida_idp.IDB_Hooks.__init__(self)

    def renamed(self, ea, new_name, local_name):
        if ida_struct.is_member_id(ea) or ida_struct.get_struc(ea) or ida_enum.get_enum_name(ea):
            return 0

        # renaming a function header
        ida_func = idaapi.get_func(ea)
        if ida_func and ida_func.start_ea == ea:
            decomp_server.cache["function_headers"][str(ea)]["name"] = new_name
            return 0

        # assume we are renaming a global var of some sort
        try:
            decomp_server.cache["global_vars"][str(ea)]["name"] = new_name
        except KeyError:
            # okay its not a global
            pass

        return 0


#
# UI
#

class ConfigDialog(QDialog):
    def __init__(self, change_hook, parent=None):
        super().__init__(parent)
        self.change_hook = change_hook
        self.setWindowTitle("Configure Decomp2DBG")
        self._main_layout = QVBoxLayout()
        self._host_edit = None  # type:QLineEdit
        self._port_edit = None  # type:QLineEdit

        self._init_widgets()
        self.setLayout(self._main_layout)
        self.show()

    def _init_widgets(self):
        upper_layout = QGridLayout()

        host_label = QLabel(self)
        host_label.setText("Host")
        self._host_edit = QLineEdit(self)
        self._host_edit.setText("localhost")
        row = 0
        upper_layout.addWidget(host_label, row, 0)
        upper_layout.addWidget(self._host_edit, row, 1)
        row += 1

        port_label = QLabel(self)
        port_label.setText("Port")
        self._port_edit = QLineEdit(self)
        self._port_edit.setText("3662")
        upper_layout.addWidget(port_label, row, 0)
        upper_layout.addWidget(self._port_edit, row, 1)
        row += 1

        # buttons
        self._ok_button = QPushButton(self)
        self._ok_button.setText("OK")
        self._ok_button.setDefault(True)
        self._ok_button.clicked.connect(self._on_ok_clicked)
        cancel_button = QPushButton(self)
        cancel_button.setText("Cancel")
        cancel_button.clicked.connect(self._on_cancel_clicked)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self._ok_button)
        buttons_layout.addWidget(cancel_button)

        # main layout
        self._main_layout.addLayout(upper_layout)
        self._main_layout.addLayout(buttons_layout)

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        global decomp_server

        host = self._host_edit.text()
        port = self._port_edit.text()

        if not host:
            QMessageBox(self).critical(None, "Invalid host",
                                       "Host cannot be empty."
                                       )
            return

        if not port:
            QMessageBox(self).critical(None, "Invalid port",
                                       "Port cannot be empty"
                                       )
            return

        decomp_server = IDADecompilerServer()
        t = threading.Thread(target=decomp_server.start_xmlrpc_server, kwargs={'host': host, 'port': int(port)})
        t.daemon = True
        try:
            t.start()
            # start hooks on good connection
            self.change_hook.hook()
        except Exception as e:
            QMessageBox(self).critical(None, "Error starting Decomp2DBG Server", str(e))
            traceback.print_exc()
            return

        self.close()

    def _on_cancel_clicked(self):
        self.close()


#
# Action Handlers
#

class IDAActionHandler(idaapi.action_handler_t):
    def __init__(self, action, plugin, typ):
        super(IDAActionHandler, self).__init__()
        self.action = action
        self.plugin = plugin
        self.typ = typ

    def update(self, ctx):
        return idaapi.AST_ENABLE_ALWAYS


class Decomp2DBGPlugin(QObject, idaapi.plugin_t):
    """Plugin entry point. Does most of the skinning magic."""

    flags = idaapi.PLUGIN_FIX
    comment = "Syncing decompiler info to GDB"
    help = "Decomp2DBG Help"
    wanted_name = "Decomp2DBG: configure"
    wanted_hotkey = "Ctrl-Shift-D"

    def __init__(self, *args, **kwargs):
        QObject.__init__(self, *args, **kwargs)
        idaapi.plugin_t.__init__(self)
        self.change_hook = IDBHooks()

    def init(self):
        return idaapi.PLUGIN_KEEP

    def run(self, arg):
        self.open_config_dialog()

    def open_config_dialog(self):
        dialog = ConfigDialog(self.change_hook)
        dialog.exec_()

    def term(self):
        pass




import argparse

from .installer import Decomp2dbgInstaller


def main():
    parser = argparse.ArgumentParser(
        description="""
            The decomp2dbg Command Line Util. 
            """,
        epilog="""
            Examples:
            decomp2dbg --install
            """
    )
    parser.add_argument(
        "--install", action="store_true", help="""
        Install the decomp2dbg core to supported decompilers as plugins. This option will start an interactive
        prompt asking for install paths for all supported decompilers. Each install path is optional and 
        will be skipped if not path is provided during install. 
        """
    )
    args = parser.parse_args()

    if args.install:
        Decomp2dbgInstaller().install()


if __name__ == "__main__":
    main()



from ...utils import *
from .utils import pc


class DecompilerPane:
    def __init__(self, decompiler, printer=pprint):
        self.decompiler: "GDBDecompilerClient" = decompiler

        self.ready_to_display = False
        self.decomp_lines = []
        self.curr_line = -1
        self.curr_func = ""
        self.print = printer

        # XXX: this needs to be removed in the future
        self.stop_global_import = False

    def update_event(self, pc_):
        if (self.decompiler.gdb_client.base_addr_end is not None) and \
                (self.decompiler.gdb_client.base_addr_start is not None):
            if pc_ > self.decompiler.gdb_client.base_addr_end or pc_ < self.decompiler.gdb_client.base_addr_start:
                return False

        
        rebased_pc = self.decompiler.rebase_addr(pc_)

        # update all known function names
        self.decompiler.update_symbols()

        # decompile the current pc location
        try:
            resp = self.decompiler.decompile(rebased_pc)
        except Exception as e:
            warn(f"Decompiler failed to get a response from decompiler on "
                 f"{hex(rebased_pc) if isinstance(rebased_pc,int) else rebased_pc} with: {e}")
            return False

        # set the decompilation for next use in display_pane
        decompilation = resp['decompilation']
        if not decompilation:
            warn("Decompiler server sent back a response without decompilation lines for "
                 f"{hex(rebased_pc) if isinstance(rebased_pc,int) else rebased_pc}")
            return False
        self.decomp_lines = decompilation
        last_line = self.curr_line
        self.curr_line = resp["curr_line"]
        if self.curr_line == -1:
            self.curr_line = last_line if last_line is not None else 0

        self.curr_func = resp["func_name"]

        # update the data known in the function (stack variables)
        self.decompiler.update_function_data(rebased_pc)
        return True

    def display_pane(self):
        """
        Display the current decompilation, with an arrow next to the current line.
        """
        if not self.decompiler.connected:
            return

        if not self.ready_to_display:
            err("Unable to decompile function")
            return

        # configure based on source config
        past_lines_color = "gray"
        nb_lines = 6
        cur_line_color = "green"

        if len(self.decomp_lines) < nb_lines:
            nb_lines = len(self.decomp_lines)

        # use GEF source printing method
        for i in range(self.curr_line - nb_lines + 1, self.curr_line + nb_lines):
            if i < 0:
                continue

            if i < self.curr_line:
                self.print(
                    "{}".format(Color.colorify("  {:4d}\t {:s}".format(i + 1, self.decomp_lines[i], ), past_lines_color))
                )

            if i == self.curr_line:
                prefix = "{}{:4d}\t ".format(RIGHT_ARROW[1:], i + 1)
                self.print(Color.colorify("{}{:s}".format(prefix, self.decomp_lines[i]), cur_line_color))

            if i > self.curr_line:
                try:
                    self.print("  {:4d}\t {:s}".format(i + 1, self.decomp_lines[i], ))
                except IndexError:
                    break
        return

    def title(self):
        """
        Special note: this function is always called before display_pane
        """
        if not self.decompiler.connected:
            return None

        self.ready_to_display = self.update_event(pc())

        if self.ready_to_display:
            title = "decompiler:{:s}:{:s}:{:d}".format(self.decompiler.name, self.curr_func, self.curr_line+1)
        else:
            title = None

        return title

    def display_pane_and_title(self, *args, **kwargs):

        #
        # title
        #

        title_ = self.title()

        line_color = "gray"
        msg_color = "cyan"
        tty_rows, tty_columns = get_terminal_size()

        if title_ is None:
            self.print(Color.colorify(HORIZONTAL_LINE * tty_columns, line_color))
        else:
            trail_len = len(title_) + 6
            title = ""
            title += Color.colorify("{:{padd}<{width}} ".format("",
                                                                width=max(tty_columns - trail_len, 0),
                                                                padd=HORIZONTAL_LINE),
                                    line_color)
            title += Color.colorify(title_, msg_color)
            title += Color.colorify(" {:{padd}<4}".format("", padd=HORIZONTAL_LINE),
                                    line_color)
            self.print(title)

        #
        # decompilation
        #

        self.display_pane()
        self.print(Color.colorify(HORIZONTAL_LINE * tty_columns, line_color))
